/*
##############################################################################
#
# MODULE:       r.randomwalk.c
#
# AUTHOR:       Martin Mergili
# CONTRIBUTORS: Massimiliano Alvioli, Ivan Marchesini
#
# PURPOSE:      Conceptual tool for mass movement propagation
#               Routing script
#
# COPYRIGHT:    (c) 2010 - 2015 by the author
#               (c) 2010 - 2015 by the BOKU University, Vienna
#               (c) 2015        by the University of Vienna
#               (c) 2000 - 2015 by the GRASS Development Team
#
# VERSION:      20150825 (25 August 2015)
#
#               This program is free software under the GNU General Public
#               License (>=v2). Read the file COPYING that comes with GRASS
#               for details.
#
##############################################################################
*/


// LOADING LIBRARIES --------------------------------------------------------------------------


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <grass/gis.h>
#include <grass/gprojects.h>
#include <grass/glocale.h>
#include <grass/raster.h>
#include <fcntl.h>
#include <time.h>
#include <grass/segment.h>


// DECLARING GLOBAL VARIABLES AND DEFINING FUNCTIONS ------------------------------------------

int m, n, CONST_NSEGCR, CONST_NSEGS;
double CONST_UDM, CONST_UDP;

struct GModule *module;
struct Cell_head cellhd;

int **alloc_imatrix(int mrows, int mcols) { // function for allocating 2D integer arrays
    int **mm;
    int mi;
    mm = (int **) calloc(mrows, sizeof(int *));
    mm[0] = (int *) calloc(mrows * mcols, sizeof(int));
    for (mi = 1; mi < mrows; mi++)
    mm[mi] = mm[mi - 1] + mcols;
    return mm;
}

int fiparam ( FILE *gfparam ) { // function for reading integer parameters

    int gparam;
    char gsparam[25], **gstrtodhlp=NULL;
    fgets(gsparam, 25, gfparam);
    gparam=strtod(gsparam, gstrtodhlp);
    return gparam;
}

double fdparam ( FILE *gfparam ) { // function for reading double parameters

    double gparam;
    char gsparam[25], **gstrtodhlp=NULL;
    fgets(gsparam, 25, gfparam);
    gparam=strtod(gsparam, gstrtodhlp);
    return gparam;
}

char *fcparam ( FILE *gfparam ) { // function for reading character parameters

    char *gsparam = (char*) malloc ( sizeof(char) * 1000 );
    fgets(gsparam, 1000, gfparam);
    gsparam[strnlen(gsparam, 1010) - 1] = '\0';
    return gsparam;
}

SEGMENT finrasti ( char gm[100] ) { // input of integer GRASS raster maps

    int gf, gx, gy, gin;
    CELL *gc;
    SEGMENT gseg_in;

    Segment_open (&gseg_in, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(int), CONST_NSEGS);

    gc = Rast_allocate_c_buf();
    gf = Rast_open_old ( gm, G_mapset() );

    for ( gx = 0; gx < m; gx++ ) {

        Rast_get_c_row ( gf, gc, gx );

        for ( gy = 0; gy < n; gy++ ) {

            if ( !Rast_is_c_null_value ( gc + gy ) ) gin = ( int ) gc[gy];
            else gin = (int)CONST_UDM;
            Segment_put(&gseg_in, &gin, gx, gy);
	}
    }
    Segment_flush(&gseg_in);
    G_free ( gc );
    Rast_close ( gf );

    return gseg_in;
}

SEGMENT finrastd ( char gm[100] ) { // input of double GRASS raster maps

    int gf, gx, gy;
    double gin;
    DCELL *gc;
    SEGMENT gseg_in;

    Segment_open (&gseg_in, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(double), CONST_NSEGS);

    gc = Rast_allocate_d_buf();
    gf = Rast_open_old ( gm, G_mapset() );

    for ( gx = 0; gx < m; gx++ ) {

        Rast_get_d_row ( gf, gc, gx );

        for ( gy = 0; gy < n; gy++ ) {

            if ( !Rast_is_d_null_value ( gc + gy ) ) gin = ( double ) gc[gy];
            else gin = CONST_UDM;
            Segment_put(&gseg_in, &gin, gx, gy);
	}
    }

    Segment_flush(&gseg_in);
    G_free ( gc );
    Rast_close ( gf );

    return gseg_in;
}

void foutrasti ( char gm[50], SEGMENT gseg_v ) { // output of integer GRASS raster maps

    int gf, gx, gy;
    int gv;
    CELL *gc;

    if ( gm != NULL ) {
        gc = Rast_allocate_c_buf();
	gf = Rast_open_fp_new( gm );
    }

    for ( gx = 0; gx < m; gx++ ) {

        if ( gm != NULL ) {
	    for ( gy = 0; gy < n; gy++ ) {
                Segment_get(&gseg_v, &gv, gx, gy);
                gc[gy] = (CELL) gv;
            }
	    Rast_put_c_row( gf, gc );
        }
    }
    G_free( gc );
    Rast_close( gf );

    return;
}

void foutrastd ( char gm[50], SEGMENT gseg_v ) { // output of double GRASS raster maps

    int gf, gx, gy;
    double gv;
    DCELL *gc;

    if ( gm != NULL ) {
        gc = Rast_allocate_d_buf();
	gf = Rast_open_fp_new( gm );
    }

    for ( gx = 0; gx < m; gx++ ) {

        if ( gm != NULL ) {
	    for ( gy = 0; gy < n; gy++ ) {
                Segment_get(&gseg_v, &gv, gx, gy);
                gc[gy] = (DCELL) gv;
            }
	    Rast_put_d_row( gf, gc );
        }
    }
    G_free( gc );
    Rast_close( gf );

    return;
}

double falpha ( double gbetax, double gbetay ) { // function for flow direction

    //int gfdir;
    double galpha;
    double GPI = 3.14159;

    if ( gbetax == 0 && gbetay == 0 ) galpha = 0;
    else if ( gbetax == 0 && gbetay > 0 ) galpha = 90;
    else if ( gbetax == 0 && gbetay < 0 ) galpha = 270;
    else if ( gbetax >= 0 ) galpha = atan( tan( gbetay ) / tan( gbetax )) * 180 / GPI;
    else if ( gbetax < 0 ) galpha = atan( tan( gbetay ) / tan( gbetax )) * 180 / GPI + 180;

    if ( galpha < 0 ) galpha += 360;
    if ( galpha > 360 ) galpha -= 360;

    /*if      (                   galpha <=  22.5 ) gfdir = 6;
    else if ( galpha >  22.5 && galpha <=  67.5 ) gfdir = 7;
    else if ( galpha >  67.5 && galpha <= 112.5 ) gfdir = 8;
    else if ( galpha > 112.5 && galpha <= 157.5 ) gfdir = 1;
    else if ( galpha > 157.5 && galpha <= 202.5 ) gfdir = 2;
    else if ( galpha > 202.5 && galpha <= 247.5 ) gfdir = 3;
    else if ( galpha > 247.5 && galpha <= 292.5 ) gfdir = 4;
    else if ( galpha > 292.5 && galpha <= 337.5 ) gfdir = 5;
    else if ( galpha > 337.5                    ) gfdir = 6;*/

    return galpha;
    //return gfdir;
}


// CALLING MAIN FUNCTION ----------------------------------------------------------------------


int main(int argc, char *argv[]) {


// DECLARING VARIABLES ------------------------------------------------------------------------


    int PAR_AFLAG, PAR_MULT, PAR_BACK, PAR_SWALK, PAR_CTRLRELEASEMAP, PAR_CTRLPROBMAP, PAR_CTRLSCOREMAP, PAR_CTRLIMPOBJ,
        PAR_NIMPOBJ, PAR_CTRLOBJSCORES, PAR_NCASES, PAR_NCASETYPES, PAR_NMODELS, PAR_NRUNS, **PAR_CASERULE, PAR_SCORE, PAR_PROB,
        PAR_X, PAR_CTRLCDF, *PAR_MID, *PAR_MTYPE, *CASE_NAME, *CASE_TYPE, *CASE_S, *IMP_NAME, *IMP_TYPE, *IMP_S, *CDF_O, SUBSET_MIN,
        SUBSET_MAX, *ris_obj, i, j, k, l, t, u, x, y, caseid, ncase, ncasetype, casetype, model, model_min, model_max, nimpobj, imparea,
        xint, releasemap, releasemap2, casemax, ncases, cdfid, ctrl_aoi;

    double CONST_PI, PAR_NRUNS0, PAR_SMIN, *PAR_RUNUPMAX, *PAR_FSLOPE, *PAR_FDIR, *PAR_SEG, *PAR_SCTRL, *PAR_MPAR1, *PAR_MPAR2, *PAR_MPAR3, *CASE_V,
        *CASE_Q, *CASE_PROB, *CASE_XR, *CASE_YR, *CASE_XS, *CASE_YS, *CDF, csz, bd_west, bd_north, deltas, prob;

    FILE *fparam, *fcases, *fobjscores, *fcdf, *fsubset;
    SEGMENT seg_aoi, seg_elev, seg_releasemap, seg_releasemap2, seg_impobj, seg_imparea, seg_scoremap, seg_probmap;
    time_t timer;

    char *path = (char*) malloc( sizeof(char) * 500 );
    char *pathd = (char*) malloc( sizeof(char) * 500 );
    char *paths = (char*) malloc( sizeof(char) * 500 );
    char *patho = (char*) malloc( sizeof(char) * 500 );
    char *wkdir = (char*) malloc( sizeof(char) * 500 );
    char *xchar = (char*) malloc( sizeof(char) * 500 );
    char *yint = (char*) malloc( sizeof(char) * 100 );

    char *PAR_PREFIX = (char*) malloc( sizeof(char) * 100 );
    char *PAR_FILEPATH = (char*) malloc( sizeof(char) * 500 );
    char *PAR_AOI = (char*) malloc( sizeof(char) * 100 );
    char *PAR_ELEV = (char*) malloc( sizeof(char) * 100 );
    char *PAR_RELEASEMAP = (char*) malloc( sizeof(char) * 100 );
    char *PAR_IMPOBJ = (char*) malloc( sizeof(char) * 100 );
    char *PAR_IMPAREA = (char*) malloc( sizeof(char) * 100 );
    char *PAR_SCOREMAP = (char*) malloc( sizeof(char) * 100 );
    char *PAR_PROBMAP = (char*) malloc( sizeof(char) * 100 );


// SETTING CONSTANT PARAMETERS ----------------------------------------------------------------


    CONST_UDM = -9999.0;
    CONST_UDP = 9999.0;
    CONST_NSEGS = 16;
    CONST_NSEGCR = 16;
    CONST_PI = 3.14159;


// INITIALIZING GIS FUNCTIONS -----------------------------------------------------------------


    G_gisinit( argv[0] );
    module = G_define_module();
    G_get_set_window( &cellhd );


// SETTING EXTENT AND CELL SIZE ---------------------------------------------------------------


    csz=cellhd.ew_res;
    bd_west=cellhd.west;
    bd_north=cellhd.north;
    n=cellhd.cols;
    m=cellhd.rows;


// PREPARING ENVIRONMENT ----------------------------------------------------------------------


    sprintf(wkdir, "%s/%s/.tmp", G_location_path(),G_mapset()); // defining working directory
    xint=10;
    yint=getenv("XINT");
    xint=atoi(yint);
    G_init_tempfile(); // initializing generation of temporary files


// READING PARAMETER FILE ---------------------------------------------------------------------


    sprintf(path, "%s/rtemp/param%d.txt", wkdir, xint); // writing name of parameter file to string
    fparam=fopen(path, "r"); // opening parameter file

    PAR_AFLAG = fiparam ( fparam ); // control for average of impact probability
    PAR_BACK = fiparam ( fparam ); // control for back-calculation of observation
    PAR_MULT = fiparam ( fparam ); // control for multiple model runs
    PAR_PROB = fiparam ( fparam ); // control for probabilities
    PAR_SCORE = fiparam ( fparam ); // control for scores
    PAR_SWALK = fiparam ( fparam ); // control for separate random walks for each model
    PAR_X = fiparam ( fparam ); // control for release from pixels

    PAR_PREFIX = fcparam ( fparam ); // prefix for output
    PAR_FILEPATH = fcparam ( fparam ); // path for output files

    PAR_AOI = fcparam ( fparam ); // area of interest raster map
    seg_aoi = finrasti( PAR_AOI );

    PAR_ELEV = fcparam ( fparam ); // elevation raster map
    seg_elev = finrastd( PAR_ELEV );

    PAR_CTRLPROBMAP = fiparam ( fparam ); // control for release probability raster map
    if ( PAR_CTRLPROBMAP == 1 ) {
        PAR_PROBMAP = fcparam ( fparam ); // release probability raster map
        seg_probmap = finrastd( PAR_PROBMAP );
    }

    PAR_CTRLRELEASEMAP = fiparam ( fparam ); // control for release area raster map
    if ( PAR_CTRLRELEASEMAP == 1 ) {
        PAR_RELEASEMAP = fcparam ( fparam ); // release area raster map
        seg_releasemap = finrasti( PAR_RELEASEMAP );
    }
    if ( PAR_X == 1 ) Segment_open (&seg_releasemap2, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(int), CONST_NSEGS);

    PAR_CTRLSCOREMAP = fiparam ( fparam ); // control for release hazard indicator score raster map
    if ( PAR_CTRLSCOREMAP == 1 ) {
        PAR_SCOREMAP = fcparam ( fparam ); // release hazard indicator score raster map
        seg_scoremap = finrastd( PAR_SCOREMAP );
    }

    PAR_CTRLIMPOBJ = fiparam ( fparam ); // control for impact objects raster map
    if ( PAR_CTRLIMPOBJ == 1 ) {
        PAR_IMPOBJ = fcparam ( fparam ); // impact objects raster map
        seg_impobj = finrasti( PAR_IMPOBJ );
    }

    PAR_CTRLOBJSCORES = fiparam ( fparam ); // control for object scores file
    if ( PAR_CTRLOBJSCORES == 1 ) PAR_NIMPOBJ = fiparam ( fparam ); // number of impact objects

    if ( PAR_BACK == 1 ) {
        PAR_IMPAREA = fcparam ( fparam ); // impact area raster map
        seg_imparea = finrasti( PAR_IMPAREA );
    }

    PAR_NCASES = fiparam ( fparam ); // number of cases
    PAR_NMODELS = fiparam ( fparam ); // number of models
    PAR_NRUNS = pow( 10, fdparam ( fparam )); // number of random walks for each case
    PAR_SMIN = fdparam ( fparam ); // minimum travel distance

    PAR_RUNUPMAX = (double*) malloc( sizeof(double) * PAR_NMODELS );
    PAR_FSLOPE = (double*) malloc( sizeof(double) * PAR_NMODELS );
    PAR_FDIR = (double*) malloc( sizeof(double) * PAR_NMODELS );
    PAR_SEG = (double*) malloc( sizeof(double) * PAR_NMODELS );
    PAR_SCTRL = (double*) malloc( sizeof(double) * PAR_NMODELS );


    if ( PAR_SWALK == 0 ) {

        PAR_SCTRL[0] = fdparam ( fparam ); // control distance for maintenance of direction
        PAR_SEG[0] = fdparam ( fparam ); // segment length
        PAR_RUNUPMAX[0] = fdparam ( fparam ); // maximum runup
        PAR_FSLOPE[0] = fdparam ( fparam ); // factor for weighting slope
        PAR_FDIR[0] = fdparam ( fparam ); // factor for weighing maintenance of direction
    }

    PAR_MID = (int*) malloc( sizeof(int) * PAR_NMODELS );
    PAR_MTYPE = (int*) malloc( sizeof(int) * PAR_NMODELS );
    PAR_MPAR1 = (double*) malloc( sizeof(double) * PAR_NMODELS );
    PAR_MPAR2 = (double*) malloc( sizeof(double) * PAR_NMODELS );
    PAR_MPAR3 = (double*) malloc( sizeof(double) * PAR_NMODELS );

    for ( model = 0; model < PAR_NMODELS; model++ ) { // loop over all models:

        PAR_MID[model] = fiparam ( fparam ); // model id
        PAR_MTYPE[model] = fiparam ( fparam ); // type of model
        PAR_MPAR1[model] = fdparam ( fparam ); // first model parameter
        PAR_MPAR2[model] = fdparam ( fparam ); // second model parameter
        PAR_MPAR3[model] = fdparam ( fparam ); // third model parameter

        if ( PAR_SWALK == 1 ) {

            PAR_SCTRL[model] = fdparam ( fparam ); // control distance for maintenance of direction
            PAR_SEG[model] = fdparam ( fparam ); // segment length
            PAR_RUNUPMAX[model] = fdparam ( fparam ); // maximum runup
            PAR_FSLOPE[model] = fdparam ( fparam ); // factor for weighting slope
            PAR_FDIR[model] = fdparam ( fparam ); // factor for weighing maintenance of direction
        }
    }

    PAR_NCASETYPES = fiparam ( fparam ); // number of case types
    if ( PAR_NCASETYPES > 0 ) {

        PAR_CASERULE = alloc_imatrix( PAR_NCASETYPES+1, PAR_NMODELS );
        for ( ncasetype = 0; ncasetype < PAR_NCASETYPES; ncasetype++ ) { // loop over all case types:

            casetype = fiparam ( fparam ); // case type
            for ( model = 0; model < PAR_NMODELS; model++ ) PAR_CASERULE[casetype][model] = fdparam ( fparam ); // case rule for model
        }
    }

    PAR_CTRLCDF = fiparam ( fparam ); // control for cdf file
    fclose(fparam); // closing parameter file


// READING CDF, SUBSETTING, CASE AND OBJECT SCORES FILES --------------------------------------


    if ( PAR_CTRLCDF == 1 ) { // file with cumulative density function:

        CDF_O = (int*) malloc( sizeof(int) * 601 );
        CDF = (double*) malloc( sizeof(double) * 601 );

        if ( PAR_MULT == 0 ) sprintf(path, "%s/rtemp/cdf.txt", wkdir); // writing name of cdf file to string
        else sprintf(path, "%s/rtemp/cdf%i.txt", wkdir, xint);
        fcdf=fopen(path, "r"); // opening cdf file

        for ( cdfid = 0; cdfid < 601; cdfid++ ) { // loop over all cdf values:

            CDF_O[cdfid] = (int)(10 * fdparam ( fcdf )); // angle of reach
            CDF[cdfid] = fdparam ( fcdf ); // cdf
        }
        fclose(fcdf); // closing cdf file
    }

    if ( PAR_MULT == 1 && PAR_PROB == 1 && PAR_NCASES > 0 ) { // file for subsetting case file:

        sprintf(path, "%s/rtemp/cdfthr%i.txt", wkdir, xint);
        fsubset=fopen(path, "r"); // opening file with subset parameters

        SUBSET_MIN = fiparam ( fsubset ); // first case to apply
        SUBSET_MAX = fiparam ( fsubset ); // last case to apply

        fclose(fsubset); // closing cdf file
    }
    else { // if file is not provided, applying entire set of cases:
        SUBSET_MIN = 0; // first case to apply
        SUBSET_MAX = 999999; // last case to apply
    }

    if ( PAR_NCASES > 0 ) { // if release file is provided, reading release file:

        ncases = PAR_NCASES + 1; // applying number of cases from parameter input
        sprintf(path, "%s/rtemp/releasefile.txt", wkdir); // writing name of cases file to string
        fcases=fopen(path, "r"); // opening cases file

        CASE_NAME = (int*) malloc( sizeof(int) * ncases );
        CASE_TYPE = (int*) malloc( sizeof(int) * ncases );
        CASE_V = (double*) malloc( sizeof(double) * ncases );
        CASE_Q = (double*) malloc( sizeof(double) * ncases );
        CASE_S = (int*) malloc( sizeof(int) * ncases );
        CASE_PROB = (double*) malloc( sizeof(double) * ncases );
        CASE_XR = (double*) malloc( sizeof(double) * ncases );
        CASE_YR = (double*) malloc( sizeof(double) * ncases );
        CASE_XS = (double*) malloc( sizeof(double) * ncases );
        CASE_YS = (double*) malloc( sizeof(double) * ncases );

        for ( caseid = 1; caseid < ncases; caseid++ ) { // loop over all cases:

            CASE_NAME[caseid] = fiparam ( fcases ); // case name
            CASE_TYPE[caseid] = fiparam ( fcases ); // case type
            CASE_V[caseid] = fdparam ( fcases ); // case magnitude
            CASE_Q[caseid] = fdparam ( fcases ); // case peak discharge
            CASE_S[caseid] = fiparam ( fcases ); // case release hazard indicator score
            CASE_PROB[caseid] = fdparam ( fcases ); // case release probability
            CASE_XR[caseid] = fdparam ( fcases ); // x coordinate of release point
            CASE_YR[caseid] = fdparam ( fcases ); // y coordinate of release point
            CASE_XS[caseid] = fdparam ( fcases ); // x coordinate of starting point
            CASE_YS[caseid] = fdparam ( fcases ); // y coordinate of starting point
        }
        fclose(fcases); // closing case file
    }
    else { // if no case file is provided:

        ncases = 0; // initializing number of cases
        caseid = 0; // initializing case id
        for ( x=0; x<m; x++ ) { // loop over all pixels:
            for ( y=0; y<n; y++ ) {

                if ( PAR_CTRLPROBMAP == 0 && PAR_BACK == 0 )

                    Segment_get(&seg_releasemap, &releasemap2, x, y ); // release map

                else { // if release probability is defined, assigning separate case ids to each pixel within area of interest:

                    Segment_get(&seg_aoi, &ctrl_aoi, x, y ); // area of interest
                    if ( PAR_BACK == 1 ) Segment_get(&seg_releasemap, &releasemap2, x, y ); // release map

                    if (( PAR_BACK == 0 && ctrl_aoi > 0 ) || ( PAR_BACK == 1 && releasemap2 > 0 )) {
                        caseid += 1; // increasing case id
                        releasemap2 = caseid; 
                    }
                    else releasemap2 = CONST_UDM;
                }

                Segment_put(&seg_releasemap2, &releasemap2, x, y ); // applying case id to pixel of release map
                if ( releasemap2 > ncases ) ncases = releasemap2; // updating number of cases
            }
        }
        Segment_flush(&seg_releasemap2);
        ncases += 1; // finalizing number of cases
    }

    if ( PAR_CTRLOBJSCORES == 1 ) { // object scores file:

        IMP_NAME = (int*) malloc( sizeof(int) * PAR_NIMPOBJ );
        IMP_TYPE = (int*) malloc( sizeof(int) * PAR_NIMPOBJ );
        IMP_S = (int*) malloc( sizeof(int) * PAR_NIMPOBJ );
        ris_obj = (int*) malloc( sizeof(int) * PAR_NIMPOBJ );

        sprintf(path, "%s/rtemp/objscores.txt", wkdir); // writing name of object scores file to string
        fobjscores=fopen(path, "r"); // opening object scores file

        for ( nimpobj = 0; nimpobj < PAR_NIMPOBJ; nimpobj++ ) {

            IMP_NAME[nimpobj] = fiparam ( fobjscores ); // name of impact object
            IMP_TYPE[nimpobj] = fiparam ( fobjscores ); // type of impact object
            IMP_S[nimpobj] = fiparam ( fobjscores ); // object score
            ris_obj[nimpobj] = 0;
        }
        fclose(fobjscores); // closing object scores file
    }


// DECLARING WORKING AND OUTPUT VARIABLES -----------------------------------------------------


    int kimpact, ctr_run, ctr_run_test, xstart, ystart, xf, yf, xseg, yseg, ctr_failalloc, fdir, randmax,
        *impquant, impquantold, impquantnew, ctrl_M[PAR_NMODELS], fdir_prev, ctrl_valid, ctrl_valid2, ctrl_jimpact, iimpact, jimpact,
        ctrl_releasemap, ctrl_impobj, ctrl_impbar, ctrlall, his_test, *his, ris, iisold, iMold, iMnew, hisold, xstop, ystop, xrun, yrun,
        ctrl_run, kall, ctr_kall, ctr_models, ctr_rscore, rid, xprob, yprob, xstart2, ystart2, aoi_f, xprob_test, yprob_test,
        casename, casename3, cases, nimpact_all, fdirtest;

    double pcomp, elev, elev_release, elev_start, elev_f, elev_segstart, elev_segstart0, randnum, elev_adj, elev_ctrl_prev, z_mov, runupmax,
        direction_factor, dist_ctrl_app, dist_ctrl_prevapp, z_movmax, smov_test, smov_outold, zmov_outold, omega_outold, *s_mov_out,
        *z_mov_out, *omega_out, omega_det, s_mov_det, s_mov_segstart0, xdalpha, xv, xv1, xv2, xzeta, xeta,
        impprob_test, *impprob, *impprobr, impprobr_test, impprobrold,impprobrnew,  impprobold, s_mov_seg, s_mov_segstart,
        relprobrnew, relprobrold, casev, caseq, caseprob, betax, betay, alphaprev, alphadiff1, alphadiff2, alphadiff;

    FILE *f_summary, *f_backfile, *f_impobj, *f_nout;
    SEGMENT seg_iM[PAR_NMODELS], seg_iis, seg_smov, seg_zmov, seg_omega, seg_his, seg_ris_impobj, seg_impprob, seg_relprobr,
        seg_impprobr, seg_impquant;

    char *mout = (char*) malloc( sizeof(char) * 50 );
    char *miis = (char*) malloc( sizeof(char) * 50 );
    char *msmov = (char*) malloc( sizeof(char) * 50 );
    char *mzmov = (char*) malloc( sizeof(char) * 50 );
    char *momega = (char*) malloc( sizeof(char) * 50 );
    char *mhis = (char*) malloc( sizeof(char) * 50 );
    char *mris_impobj = (char*) malloc( sizeof(char) * 50 );
    char *mimpprobr = (char*) malloc( sizeof(char) * 50 );
    char *mrelprobr = (char*) malloc( sizeof(char) * 50 );
    char *mimpquant = (char*) malloc( sizeof(char) * 50 );

    int *xadj = (int*) malloc( sizeof(int) * 9 );
    int *yadj = (int*) malloc( sizeof(int) * 9 );
    //int *fdirtestd = (int*) malloc( sizeof(int) * 9 );
    double *alphatestd = (double*) malloc( sizeof(double) * 9 );

    int *whtslope = (int*) malloc( sizeof(int) * 9 );
    double *xalpha = (double*) malloc( sizeof(double) * 2 );

    double *s_mov = (double*) malloc( sizeof(double) * 10000 );
    double *omega = (double*) malloc( sizeof(double) * 10000 );

    double *s_movmax = (double*) malloc( sizeof(double) * PAR_NMODELS );
    double *omega_stop = (double*) malloc( sizeof(double) * PAR_NMODELS );
    double *omega_test = (double*) malloc( sizeof(double) * PAR_NMODELS );

    int *xf_run = (int*) malloc( sizeof(int) * 10000 );
    int *yf_run = (int*) malloc( sizeof(int) * 10000 );

    int *ximpact = (int*) malloc( sizeof(int) * 100000 );
    int *yimpact = (int*) malloc( sizeof(int) * 100000 );
    int *iis = (int*) malloc( sizeof(int) * 100000 );
    int **iM = alloc_imatrix( PAR_NMODELS, 100000 );

    s_mov_out = (double*) malloc( sizeof(double) * 100000 );
    z_mov_out = (double*) malloc( sizeof(double) * 100000 );
    omega_out = (double*) malloc( sizeof(double) * 100000 );
    impquant = (int*) malloc( sizeof(int) * 100000 );

    if ( PAR_PROB == 1 ) {
        impprobr = (double*) malloc( sizeof(double) * 100000 );
        impprob = (double*) malloc( sizeof(double) * 100000 );
    }
    if ( PAR_SCORE == 1 ) his = (int*) malloc( sizeof(int) * 100000 );


// PREPARING OUTPUT FILES ---------------------------------------------------------------------


    if ( PAR_MULT == 0 ) {
        sprintf(paths, "%s%s_summary.txt", PAR_FILEPATH, PAR_PREFIX); // summary file
        if ( PAR_CTRLOBJSCORES == 1 ) sprintf(patho, "%s%s_ris.txt", PAR_FILEPATH, PAR_PREFIX); // risk indicator score file
    }
    else sprintf(paths, "%s%s_summary%d.txt", PAR_FILEPATH, PAR_PREFIX, xint); // for multiple model runs, only summary file is relevant
    if ( PAR_BACK == 1 ) sprintf(pathd, "%s%s_backfile.txt", PAR_FILEPATH, PAR_PREFIX); // back-calculation file

    f_summary=fopen(paths, "w"); // opening files:
    if ( PAR_BACK == 1 ) f_backfile=fopen(pathd, "w");
    if ( PAR_CTRLIMPOBJ == 1 && PAR_CTRLOBJSCORES == 1 ) f_impobj=fopen(patho, "w");

    if ( PAR_MULT == 1 ) {
        sprintf(path, "%s%s_parout%d.txt", PAR_FILEPATH, PAR_PREFIX, xint); // file with information on model run:
        f_nout=fopen(path, "w");
    }


// PREPARING SEGMENT FILES FOR OUTPUT RASTER MAPS ---------------------------------------------


    Segment_open (&seg_iis, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(int), CONST_NSEGS);
    Segment_open (&seg_impquant, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(int), CONST_NSEGS);

    if ( PAR_MULT == 0 ) {

        for ( model = 0; model < PAR_NMODELS; model++ )
            Segment_open (&seg_iM[model], G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(int), CONST_NSEGS);

        Segment_open (&seg_smov, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(double), CONST_NSEGS);
        Segment_open (&seg_zmov, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(double), CONST_NSEGS);
        Segment_open (&seg_omega, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(double), CONST_NSEGS);

        if ( PAR_CTRLIMPOBJ == 1 )
            Segment_open (&seg_ris_impobj, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(int), CONST_NSEGS);
    }

    if ( PAR_PROB == 1 ) {

        if ( PAR_MULT == 0 ) Segment_open (&seg_relprobr, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(double), CONST_NSEGS);
        Segment_open (&seg_impprobr, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(double), CONST_NSEGS);
        if ( PAR_MULT == 0 ) Segment_open (&seg_impprob, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(double), CONST_NSEGS);
    }

    if ( PAR_SCORE == 1 ) Segment_open (&seg_his, G_tempfile(), m, n, CONST_NSEGCR, CONST_NSEGCR, sizeof(int), CONST_NSEGS);


// PRINTING HEADERS OF OUTPUT FILES AND DISPLAY -----------------------------------------------


    fprintf(f_summary, "ID");
    printf("\n   ID\tSTATUS");

    for ( model = 0; model < PAR_NMODELS; model++ ) {
        fprintf(f_summary, "\tLMAX%i\tOMEGAT%i", PAR_MID[model], PAR_MID[model]);
        printf("\tLMAX%i\tOMEGAT%i", PAR_MID[model], PAR_MID[model]);
    }
    fprintf(f_summary, "\tRMAX\tAREA\n");
    printf("\tRMAX\tAREA\n");

    if ( PAR_BACK == 1 ) fprintf(f_backfile, "ID\tCASE\tLMAX\tOMEGAT\n");


// PREPARING ROUTING --------------------------------------------------------------------------


    timer=time(NULL); // timer for randomization
    srand(timer); // seed for random number

    rid = 0; // initializing counter for reconstruction cases
    xprob = 0; yprob = -1; // initializing start coordinates if all valid pixels are considered for release

    for ( caseid = 1; caseid < ncases; caseid++ ) { // loop over all cases:

        impquant[0] = 0; // initializing number of cases possibly affecting the pixel
        if ( PAR_PROB == 1 ) impprobr[0] = 0;  // inizializing impact probability
        jimpact = 0; // resetting counter for number of affected pixels
        ctr_kall = 0; // initializing counter for the total number of random walks for the case
        ctr_failalloc = 0; // resetting counter for unsuccessful routing attempts
        runupmax = 0; // resetting maximum runup height

        for ( model = 0; model < PAR_NMODELS; model++ ) {
            s_movmax[model] = CONST_UDM; omega_stop[model] = CONST_UDP; // resetting maximum travel distance and angle of reach
        }

        if ( PAR_NCASES > 0 ) {
            casename = CASE_NAME[caseid]; // case name
            casetype = CASE_TYPE[caseid]; // case type
            casev = CASE_V[caseid]; // case volume
            caseq = CASE_Q[caseid]; // case peak discharge
            cases = CASE_S[caseid]; // case release score
            caseprob = CASE_PROB[caseid]; // case release probability
        }
        else {
            casename = caseid; // case name
            casetype = 1; // case type
            casev = CONST_UDM; // case volume
            caseq = CONST_UDM; // case peak discharge
            cases = CONST_UDM; // case score
            caseprob = CONST_UDM; // case release probability
        }

        if ( PAR_X == 1 && PAR_CTRLRELEASEMAP == 1 && PAR_CTRLPROBMAP == 0 && PAR_BACK == 0 ) {
            // for start from pixels defined by the release map:

            kall = 0; // initializing total number of random walks for the case
            xstart2 = 0; ystart2 = 0; xstop = m - 1; ystop = n - 1; // initializing release variables

            for ( xrun=1; xrun<xstop; xrun++ ) { // loop over all pixels:
                for ( yrun=1; yrun<ystop; yrun++ ) {
                    Segment_get(&seg_releasemap2, &releasemap2, xrun, yrun); // release map (case id)
                    if ( releasemap2 == casename ) kall += PAR_NRUNS; // updating total number of random walks for the case
                }
            }
        }

        else if ( PAR_X == 1 ) { // if separate case ids were assigned to each pixel within area of interest:

            releasemap2 = CONST_UDM; // initializing case id
            while ( releasemap2 != casename && xprob < m-2 ) { // searching for the next valid pixel:

                if ( yprob > n-2 ) { yprob = 0; xprob += 1; } else yprob += 1; // updating test variables
                Segment_get(&seg_releasemap2, &releasemap2, xprob, yprob); // case id
            }
            xstart2 = xprob; ystart2 = yprob; xstop = xprob + 1; ystop = yprob + 1; // initializing release variables
            kall = PAR_NRUNS; // total number of random walks for the case
        }

        else { xstart2 = 0; ystart2 = 0; xstop = 1; ystop = 1; kall = PAR_NRUNS; }
            // release variables and total number of random walks for the case if a release file is provided

        for ( xrun=xstart2; xrun<xstop; xrun++ ) { for ( yrun=ystart2; yrun<ystop; yrun++ ) { // loop over all release pixels of the case:

            if ( PAR_X == 1 ) { // for release from pixels:

                Segment_get(&seg_releasemap2, &releasemap2, xrun, yrun); // release map (case id)
                if ( releasemap2 == casename && releasemap2 != CONST_UDM ) { // if pixel is valid:

                    ctrl_run = 1; // setting control for starting random walks to positive
                    x = xrun; y = yrun; // coordinates of release point
                    xstart = xrun; ystart = yrun; // coordinates of start of random walk
                }
                else ctrl_run = 0; // if pixel is invalid, setting control for starting random walks to negative
            }

            else { // for release from coordinates defined by release file:

                ctrl_run = 1;

                x = (int)(( bd_north - CASE_YR[caseid] ) / csz + 0.5 ); // coordinates of release point
                y = (int)(( CASE_XR[caseid] - bd_west ) / csz + 0.5 );

                xstart = (int)(( bd_north - CASE_YS[caseid] ) / csz + 0.5 ); // coordinates of start of random walk
                ystart = (int)(( CASE_XS[caseid] - bd_west ) / csz + 0.5 );
            }

            if ( PAR_NCASES > 0 ) { // if release file is provided:
                if ( PAR_PROB == 1 ) prob = caseprob; else prob = 1; // release probability
                if ( PAR_SCORE == 1 ) ctr_rscore = cases; else ctr_rscore = 1; // release hazard indicator score
                if ( SUBSET_MAX > SUBSET_MIN && ( casename >= SUBSET_MIN && casename <= SUBSET_MAX )) ctrl_run = 1;
                else if ( SUBSET_MAX < SUBSET_MIN && ( casename >= SUBSET_MIN || casename <= SUBSET_MAX )) ctrl_run = 1;
                else ctrl_run = 0; // verifying that case is in subset
            }
            else { // if no release file is provided:
                if ( PAR_CTRLPROBMAP == 1 ) Segment_get(&seg_probmap, &prob, xrun, yrun); else prob = 1;// release probability
                if ( PAR_SCORE == 1 ) Segment_get(&seg_scoremap, &ctr_rscore, xrun, yrun); else ctr_rscore = 1;
                    // release hazard indicator score
            }

            if ( PAR_BACK == 1 ) { // for back-calculation:
                if ( PAR_NCASES > 0 ) casename3 = casename;
                else Segment_get(&seg_releasemap, &casename3, xrun, yrun); // release map (case id)
            }

            if ( ctrl_run == 1 ) Segment_get(&seg_aoi, &ctrl_aoi, x, y); else ctrl_aoi = CONST_UDM; // area of interest

            if ( ctrl_run == 1 && ( PAR_X == 0 || prob > 0 || ctr_rscore > 0 )
                && x >= 2 && x <= (m-3) && y >= 2 && y <= (n-3) && ctrl_aoi > 0 ) { // if release pixel supports routing:

                Segment_get(&seg_elev, &elev_release, x, y); // elevation of release point
                Segment_get(&seg_elev, &elev_start, xstart, ystart); // elevation of start of random walk

                if ( PAR_BACK == 1  ) { omega_det = CONST_UDP; s_mov_det = CONST_UDM; }
                    // resetting travel distance and angle of reach with respect to release pixel


            // STARTING SERIES OF RANDOM WALKS ------------------------------------------------


                for ( k=0; k<PAR_NRUNS; k++ ) { // loop over defined number of random walks:

                    ctr_models = 0;
                    for ( model = 0; model < PAR_NMODELS; model++ ) { // activating models:

                        if ( PAR_NCASETYPES == 0 ) { // if no case rules are provided, activating all models
                            ctrl_M[model] = 1;
                            ctr_models += 1;
                        }
                        else if ( PAR_CASERULE[casetype][model] > 0 ) { // activating model according to case rule
                            ctrl_M[model] = 1;
                            ctr_models += 1;
                        }
                        else ctrl_M[model] = 0; // deactivating model according to case rule
                    }

                    ctr_kall+=1; // updating counter for number of random walks for case
                    pcomp = (double)100 * (double)ctr_kall / (double)kall; // per cent of case completed
                    if ( (int)pcomp % 1 == 0 ) printf ("   %i\t%.0f%%\r", casename, pcomp);

                    fdir_prev = CONST_UDM; // resetting previous flow direction
                    //alphaprev = CONST_UDM; // resetting previous flow direction
                    xf_run[0] = x; yf_run[0] = y; // setting initial coordinates to the release coordinates 
                    ximpact[0] = xstart; yimpact[0] = ystart; // storing start coordinates
                    s_mov[0] = pow ( pow ( xstart - x, 2 ) * csz + pow ( ystart - y, 2 ) * csz, 0.5 ); // initializing travel distance
                    if ( PAR_PROB == 1 ) impprobr[0] = 1; // setting impact probability
                    impquant[0] = 1; // setting number of cases possibly affecting the pixel
                    z_mov = elev_release - elev_start; // initializing loss of elevation
                    z_movmax = 0; // resetting maximum drop of elevation
                    xf = xstart; yf = ystart; // initializing coordinates for direction
                    ctrl_aoi = 1; // resetting control for area of interest
                    ctrl_valid = 1; // resetting control for raster cell validity
                    ctrl_valid2 = 1; // resetting control for validity of all adjacent pixels
                    ctrl_impbar = 0; // resetting control for barrier
                    ctrlall = 1; // resetting number of active models
                    ctr_run = 0; // resetting number of step of random walk
                    xv1 = 0; // resetting flow velocity (only used with models of type 5)
                    xseg = xstart; yseg = ystart; // resetting segment start coordinates
                    s_mov_segstart = s_mov[0]; // resetting horizontal travel distance at start of segment
                    s_mov_segstart0 = 0; // resetting previous horizontal travel distance at start of segment
                    elev_segstart0 = elev_release; // resetting previous elevation at start of segment
                    xalpha[0] = 0; // initializing previous angle along path

                    if ( PAR_SWALK == 0 || PAR_NMODELS == 1 ) {
                        // if separate routing for each model does not apply, or if only one model is used:

                        model_min = 0; // id of first model to be applied 
                        model_max = PAR_NMODELS; // id of last model to be applied
                    }
                    else { // if separate routing applies, selecting model randomly:

                        randmax=PAR_NMODELS; // maximum for random number
                        randnum=(double)(rand()%randmax); // random number

                        model_min = (int)randnum; // id of first model to be applied 
                        model_max = (int)randnum + 1; // id of last model to be applied
                    }

                    while ( ctrlall > 0 && ctrl_valid == 1 && ctrl_impbar <= 0 && ctr_run < 10000 ) { // routing random walk:


                    // DETERMINING TARGET PIXEL -----------------------------------------------


                        ctr_run += 1; // updating number of step

                        xf_run[ctr_run] = xf; yf_run[ctr_run] = yf; // storing coordinates in array

                        ctr_run_test = ctr_run - 1; // identifying control point:
                        if ( PAR_SCTRL[model_min] >= s_mov[ctr_run-1] ) ctr_run_test = 0;
                        else {
                            while ( s_mov[ctr_run-1] - s_mov[ctr_run_test] < PAR_SCTRL[model_min] && ctr_run_test > 0 ) ctr_run_test -= 1;
                        }

                        dist_ctrl_prevapp = pow ( pow ( xf * csz - xf_run[ctr_run_test] * csz, 2 )
                            + pow ( yf * csz - yf_run[ctr_run_test] * csz, 2 ), 0.5 ); // distance to control point

                        Segment_get(&seg_elev, &elev_ctrl_prev, xf_run[ctr_run_test], yf_run[ctr_run_test]); // elevation of control point

                        xadj[1]=xf-1; yadj[1]=yf+1; // coordinates of adjacent cells
                        xadj[2]=xf-1; yadj[2]=yf;
                        xadj[3]=xf-1; yadj[3]=yf-1;
                        xadj[4]=xf; yadj[4]=yf-1;
                        xadj[5]=xf+1; yadj[5]=yf-1;
                        xadj[6]=xf+1; yadj[6]=yf;
                        xadj[7]=xf+1; yadj[7]=yf+1;
                        xadj[8]=xf; yadj[8]=yf+1;
                        whtslope[0]=0; // initializing weighted slope

                        Segment_get(&seg_elev, &elev, xf, yf); // reading elevation

                        for (fdir=1; fdir<9; fdir++) { // starting loop over all neighbouring cells:

                            if ( fdir == 1 || fdir == 3 || fdir == 5 || fdir == 7 ) deltas = csz * pow (2, 0.5);
                            else deltas = csz; // updating distance from release

                            Segment_get(&seg_elev, &elev_adj, xadj[fdir], yadj[fdir]); // elevation of neighbour pixel
                            Segment_get(&seg_aoi, &ctrl_aoi, xadj[fdir], yadj[fdir]); // control for aoi of neighbour pixel

                            dist_ctrl_app = pow ( pow ( xadj[fdir] * csz - xf_run[ctr_run_test] * csz, 2 )
                                + pow ( yadj[fdir] * csz - yf_run[ctr_run_test] * csz, 2 ), 0.5 );
                                // distance of neighbour pixel to control point

                            if ( elev_adj == CONST_UDM && ctrl_aoi <= 0 ) ctrl_valid2 = 0;
                                // checking whether elevation information is available

                            if ( dist_ctrl_app > dist_ctrl_prevapp
                                && z_movmax - ( elev_release - elev_adj ) <= PAR_RUNUPMAX[model_min] ) {
                                // if outflow to this pixel is possible:

                                if ( fdir == fdir_prev ) direction_factor = 100 * pow( PAR_FDIR[model_min], 2 );
                                else if ( fdir == fdir_prev - 1 || fdir == fdir_prev + 1
                                    || ( fdir == 1 && fdir_prev == 8) || ( fdir == 8 && fdir_prev == 1) )
                                    direction_factor = 100 * PAR_FDIR[model_min];
                                else direction_factor = 100; // weighting for maintenance of flow direction

                                if (( elev - elev_adj ) / deltas <= 2.75 ) // weighting for slope:
                                    whtslope[fdir] = whtslope[fdir-1] + direction_factor
                                        * exp(( elev - elev_adj ) / deltas * PAR_FSLOPE[model_min] );
                                else whtslope[fdir] = whtslope[fdir-1] + direction_factor * exp( 2.75 * PAR_FSLOPE[model_min] );
                                    // cumulative weighted slope

                                /*if ( xadj[fdir] != xf_run[ctr_run_test] )
                                    betax = atan ( elev_ctrl_prev - elev_adj ) / (( xadj[fdir] - xf_run[ctr_run_test] ) * csz );
                                else betax = 0;
                                if ( yadj[fdir] != yf_run[ctr_run_test] )
                                    betay = atan ( elev_ctrl_prev - elev_adj ) / (( yadj[fdir] - yf_run[ctr_run_test] ) * csz );
                                else betay = 0; // x and y slopes

				alphatestd[fdir] = CONST_PI / 180 * falpha( betax, betay ); // previous flow direction

                                if ( alphaprev == CONST_UDM ) alphaprev = alphatestd[fdir];
                                alphadiff1 = fabs( alphatestd[fdir] - alphaprev );
                                alphadiff2 = fabs( alphatestd[fdir] - alphaprev - 2 * CONST_PI );
                                if ( alphadiff1 < alphadiff2 ) alphadiff = alphadiff1; else alphadiff = alphadiff2;

                                if ( alphadiff < CONST_PI / 2 ) {

                                    if (( elev_ctrl_prev - elev_adj ) / dist_ctrl_app <= 2.75 ) // weighting
                                        whtslope[fdir] = whtslope[fdir-1] + 1000 * exp( - PAR_FDIR[model_min] * tan( alphadiff ))
                                            * exp(( elev_ctrl_prev - elev_adj ) / dist_ctrl_app * PAR_FSLOPE[model_min] );
                                    else whtslope[fdir] = whtslope[fdir-1] + 1000 * exp( - PAR_FDIR[model_min] * tan( alphadiff ))
                                            * exp( 2.75 * PAR_FSLOPE[model_min] );
                                }
                                else whtslope[fdir] = whtslope[fdir-1]; */
                            }
                            else whtslope[fdir] = whtslope[fdir-1];
                                // if outflow to this pixel is not possible, not increasing cumulative weighted slope
                            if ( fdir == 8 && whtslope[fdir] > 0 && whtslope[fdir] < 0.5 ) whtslope[fdir] = 0.5;
                                // ensuring minimum of weighted slope
                        }

                        if ( whtslope[8] > 0 ) { // randomly determining target pixel:

                            randmax=(int)(whtslope[8]+0.5); // maximum for random number
                            randnum=(double)(rand()%randmax); // random number
                            if (randnum < 0.0001) randnum=0.0001;

                            for (l=1; l<9; l++) { // starting loop over all downslope directions
                                if (randnum <= whtslope[l]) { // if random number matches direction:
                                    fdir=l; // direction
                                    //alphaprev = alphatestd[l];
                                    break; // leaving loop
                                }
                            }
                        }
                        else ctrl_valid = 0; // if no outflow is possible, setting validity to negative

                        fdir_prev = fdir;
                        if (fdir == 5 || fdir == 6 || fdir == 7 ) xf += 1; // updating coordinates
                        else if (fdir == 1 || fdir == 2 || fdir == 3 ) xf -= 1;
                        if (fdir == 3 || fdir == 4 || fdir == 5 ) yf -= 1;
                        else if (fdir == 1 || fdir == 7 || fdir == 8 ) yf += 1;

                        Segment_get(&seg_elev, &elev_f, xf, yf); // elevation of target pixel
                        Segment_get(&seg_aoi, &aoi_f, xf, yf); // control for aoi of target pixel

                        if ( ctrl_valid == 1 ) { // updating distance from release coordinates:

                            s_mov_seg = pow ( pow ( xf * csz - xseg * csz, 2 ) + pow ( yf * csz - yseg * csz, 2 ), 0.5 );
                            s_mov[ctr_run] = s_mov_segstart + s_mov_seg;

                            if ( s_mov_seg >= PAR_SEG[model_min] ) {

                                xseg = xf;
                                yseg = yf;
                                s_mov_segstart = s_mov[ctr_run];
                                s_mov_segstart0 = s_mov[ctr_run-1];
                                elev_segstart0 = elev;
                            }

                            z_mov = elev_release-elev_f; // updating total loss of elevation
                            if ( z_mov > z_movmax ) z_movmax = z_mov; // updating maximum loss of elevation
                        }

                        if ( xf < 2 || xf > (m-3) || yf < 2 || yf > (n-3) || aoi_f <= 0 || ctrl_valid2 == 0 ) {
                            // if movement would leave valid zone:

                            ctr_failalloc+=1; // updating counter for unsuccessful routing attempts
                            for ( model = model_min; model < model_max; model++ ) {
                                // setting maximum travel distance and angle of reach path to undefined:
                                s_movmax[model] = CONST_UDM;
                                omega_stop[model] = CONST_UDP;
                            }
                            break; // leaving loop
                        }

                        if ( ctrl_valid == 0 ) break; // leaving loop if no more proceeding is possible

                        if ( PAR_CTRLIMPOBJ == 1 && PAR_CTRLOBJSCORES == 1 ) { // testing whether random walk hits a barrier:

                            Segment_get(&seg_impobj, &ctrl_impobj, xf, yf); // impact objects
                            for ( nimpobj = 0; nimpobj < PAR_NIMPOBJ; nimpobj++ ) { // searching for correct object id:
                                if ( IMP_NAME[nimpobj] == ctrl_impobj && IMP_TYPE[nimpobj] == 2 ) {
                                    ctrl_impbar = 1; // if impact object is defined as a barrier, setting control to positive
                                    break; // leaving loop
                                }
                            }
                        }


                    // APPLYING MODELS --------------------------------------------------------


                        if ( PAR_BACK == 1  ) Segment_get(&seg_imparea, &imparea, xf, yf); // impact area map
                        omega[ctr_run] = 180 / CONST_PI * atan ( z_mov / s_mov[ctr_run] ); // average angle of path

                        for ( model = model_min; model < model_max; model++ ) { // loop over all models:

                            if ( ctrl_M[model] == 1 ) { // if movement is still active:

                                if ( PAR_MTYPE[model] == 1 ) // model equations:
                                    omega_test[model] = PAR_MPAR1[model];
                                else if ( PAR_MTYPE[model] == 2 )
                                    omega_test[model] =
                                        180 / CONST_PI * pow( 10, atan( PAR_MPAR1[model] * log10( casev ) + PAR_MPAR2[model] ));
                                else if ( PAR_MTYPE[model] == 3 )
                                    smov_test = PAR_MPAR1[model] * pow( casev, PAR_MPAR2[model] ) * pow( z_mov, PAR_MPAR3[model] );
                                else if ( PAR_MTYPE[model] == 4 )
                                    omega_test[model] = PAR_MPAR1[model] * pow( caseq, PAR_MPAR2[model] );
                                else if ( PAR_MTYPE[model] == 5 ) { // two-parameter friction model:

                                    xalpha[1] = atan(( elev_segstart0 - elev_f ) / ( s_mov[ctr_run] - s_mov_segstart0 )); // slope

                                    if ( s_mov_segstart0 == 0 || xalpha[1] > xalpha[0] ) xdalpha = 0;
                                    else xdalpha = xalpha[0] - xalpha[1]; // curvature

                                    xv2 = xv1 * cos( xdalpha ); // corrected flow velocity

                                    xzeta = 9.81 * ( sin( xalpha[1] ) - PAR_MPAR1[model] * cos( xalpha[1] ));
                                    xeta = - 2 * ( s_mov[ctr_run] - s_mov_segstart0 ) / cos( xalpha[1] ) / PAR_MPAR2[model];
                                    if ( xzeta * PAR_MPAR2[model] * ( 1 - exp( xeta )) + pow( xv2, 2 ) * exp( xeta ) * cos( xdalpha ) > 0 )
                                        xv = pow( xzeta * PAR_MPAR2[model] * ( 1 - exp( xeta ))
                                            + pow( xv2, 2 ) * exp( xeta ) * cos( xdalpha ), 0.5 ); // updating flow velocity
                                    else xv = 0;

                                    if ( s_mov_seg >= PAR_SEG[model_min] ) {
                                        xalpha[0] = xalpha[1]; // storing slope for next step
                                        xv1 = xv;
                                    }
                                }


                            // UPDATING RESULTS -------------------------------------------------------


                                if (((( PAR_MTYPE[model] == 5 && xv > 0 ) || ( PAR_MTYPE[model] == 3 && s_mov[ctr_run] <= smov_test )
                                    || (( PAR_MTYPE[model] == 1 || PAR_MTYPE[model] == 2 || PAR_MTYPE[model] == 4 )
                                    && omega[ctr_run] >= omega_test[model]) && PAR_BACK == 0 )
                                    || ( PAR_BACK > 0 && imparea == casename3 )) || s_mov[ctr_run] < PAR_SMIN ) { // if model is still active:
                                }
                                else ctrl_M[model] = 0; // else, terminating motion
                            }
                        }

                        ctrlall = 0; for ( model = model_min; model < model_max; model++ ) ctrlall += ctrl_M[model];
                            // number of active models

                        if ( ctrlall > 0 ) { // if at least one model is still active:

                            ctrl_jimpact = 1; // initializing control for newly affected pixel
                            for ( iimpact = 0; iimpact <= jimpact; iimpact++ ) { // loop over all cells already impacted:

                                if ( xf == ximpact[iimpact] && yf == yimpact[iimpact] ) { // testing if cell was already impacted:
                                    ctrl_jimpact = 0; // setting control for newly affected pixel to negative
                                    kimpact = iimpact; // updating index of current pixel
                                }
                            }
                            if ( ctrl_jimpact == 1 ) { // if cell was not yet impacted:
                                jimpact += 1; // increasing counter
                                kimpact = jimpact; // updating index of current pixel
                                ximpact[kimpact] = xf; // storing coordinates
                                yimpact[kimpact] = yf;
                                for ( model = 0; model < PAR_NMODELS; model++ ) iM[model][kimpact] = 0; // resetting impact frequency
                                s_mov_out[kimpact] = 0; z_mov_out[kimpact] = CONST_UDM; omega_out[kimpact] = CONST_UDM;
                                if ( PAR_PROB == 1 ) { impprobr[kimpact] = 0; impprob[kimpact] = 0; impquant[kimpact] = 0; }
                                if ( PAR_SCORE == 1 ) his[kimpact] = CONST_UDM; // resetting array elements
                            }

                            if ( z_movmax - z_mov > runupmax ) runupmax = z_movmax - z_mov; // updating maximum runup height

                            if ( s_mov_out[kimpact] == 0 || s_mov[ctr_run] < s_mov_out[kimpact] ) s_mov_out[kimpact] = s_mov[ctr_run];
                                // updating travel distance for output
                            if ( z_mov_out[kimpact] == 0 ) z_mov_out[kimpact] = z_mov; // updating loss of elevation
                            if ( omega[ctr_run] > omega_out[kimpact] ) omega_out[kimpact] = omega[ctr_run];
                                // updating average angle of path

                            impquant[kimpact] = 1; // updating number of cases possibly affecting the pixel
                            for ( model = model_min; model < model_max; model++ ) {
                                if ( ctrl_M[model] == 1 ) iM[model][kimpact] += 1; // updating impact frequency
                            }

                            if ( PAR_PROB == 1 ) { // for probabilities:

                                if ( omega[ctr_run] < 60 )
                                    impprobr_test = CDF[(int)(omega[ctr_run]*10+0.5)]; // impact probability
                                else impprobr_test = 1;
                                if ( impprobr_test > impprobr[kimpact] ) impprobr[kimpact] = impprobr_test; // updating impact probability
                            }

                            if ( PAR_SCORE == 1 ) { // for scores:

                                if ( ctr_rscore == CONST_UDM ) his_test = ctrlall; // impact hazard indication score
                                else his_test = (int)( pow( pow( (double)ctr_rscore, 1.5 )
                                    * pow ( (double)iis[iimpact], 1.5 ), (double)1 / (double)3 ));
                                if ( his_test > his[kimpact] ) his[kimpact] = his_test; // updating impact hazard indication score
                            }
                        }
                    } // stopping random walk
                } // stopping series of random walks for start pixel/coordinates
            }
        }}


    // UPDATING OUTPUT WITH RESULTS FOR THE CONSIDERED CASE -----------------------------------


        for ( iimpact = 0; iimpact <= jimpact; iimpact++ ) { // loop over all impacted pixels:

            iis[iimpact] = 0; // impact indicator score:

            for ( model = 0; model < PAR_NMODELS; model++ ) {

                if ( iM[model][iimpact] > 0 ) iis[iimpact] += 1; // updating impact indicator score
                if ( s_mov_out[iimpact] > s_movmax[model] && iM[model][iimpact] > 0 ) {
                    s_movmax[model] = s_mov_out[iimpact]; // updating maximum travel distance
                    omega_stop[model] = omega_out[iimpact]; // updating average angle of flow path
                }
            }

            Segment_get(&seg_iis, &iisold, ximpact[iimpact], yimpact[iimpact]); // reading old impact indicator score
            if ( iis[iimpact] > iisold ) { // if score exceeds the old value:
                Segment_put(&seg_iis, &iis[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating score
                Segment_flush(&seg_iis);
            }

            if ( PAR_MULT == 0 ) { // for single model run:

                for ( model = 0; model < PAR_NMODELS; model++ ) { // loop over all models:

                    Segment_get(&seg_iM[model], &iMold, ximpact[iimpact], yimpact[iimpact]); // old impact frequency
                    iMnew = iMold + iM[model][iimpact]; // updating impact frequency for the considered pixel
                    Segment_put(&seg_iM[model], &iMnew, ximpact[iimpact], yimpact[iimpact]);
                    Segment_flush(&seg_iM[model]);
                }

                Segment_get(&seg_smov, &smov_outold, ximpact[iimpact], yimpact[iimpact]); // old travel distance
                if ( smov_outold == 0 || s_mov_out[iimpact] < smov_outold ) { // if distance is smaller than the old value for the pixel:
                    Segment_put(&seg_smov, &s_mov_out[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating distance
                    Segment_flush(&seg_smov);
                }

                Segment_get(&seg_zmov, &zmov_outold, ximpact[iimpact], yimpact[iimpact]); // old loss of elevation
                if ( zmov_outold == 0 || z_mov_out[iimpact] < zmov_outold ) { // if loss is smaller than the old value for the pixel:
                    Segment_put(&seg_zmov, &z_mov_out[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating distance
                    Segment_flush(&seg_zmov);
                }

                Segment_get(&seg_omega, &omega_outold, ximpact[iimpact], yimpact[iimpact]); // old average angle of path
                if ( omega_out[iimpact] > omega_outold ) { // if angle exceeds the old value for the considered pixel:
                    Segment_put(&seg_omega, &omega_out[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating angle
                    Segment_flush(&seg_omega);
                }

                if ( PAR_SCORE == 1 ) { // hazard indicator score:

                    Segment_get(&seg_his, &hisold, ximpact[iimpact], yimpact[iimpact]); // old impact hazard indicator score
                    if ( his[iimpact] > hisold ) { // if score exceeds the old value for the considered pixel:
                        Segment_put(&seg_his, &his[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating score
                        Segment_flush(&seg_his);
                    }
                }

                if ( PAR_PROB == 1 ) { // average release probability of possible release zone:

                    Segment_get(&seg_relprobr, &relprobrold, ximpact[iimpact], yimpact[iimpact]); // old sum of release probabilities
                    relprobrnew = relprobrold + prob; // updating sum
                    if ( relprobrnew > 0 ) {
                        Segment_put(&seg_relprobr, &relprobrnew, ximpact[iimpact], yimpact[iimpact]);
                        Segment_flush(&seg_relprobr);
                    }
                }
            }

            Segment_get(&seg_impquant, &impquantold, ximpact[iimpact], yimpact[iimpact]);
                // old number of cases possibly affecting the pixel
            impquantnew = impquantold + impquant[iimpact]; // updating area
            if ( impquantnew > 0 ) {
                Segment_put(&seg_impquant, &impquantnew, ximpact[iimpact], yimpact[iimpact]);
                Segment_flush(&seg_impquant);
            }

            if ( PAR_PROB == 1 ) { // for probabilities:

                Segment_get(&seg_impprobr, &impprobrold, ximpact[iimpact], yimpact[iimpact]); // old (sum of) impact probability

                if ( PAR_AFLAG == 0 ) {
                    if ( impprobr[iimpact] > impprobrold ) impprobrnew = impprobr[iimpact]; // updating impact probability:
                    else impprobrnew = 0;
                }
                else impprobrnew = impprobrold + impprobr[iimpact]; // updating sum of impact probability

                if ( impprobrnew > 0 ) {
                    Segment_put(&seg_impprobr, &impprobrnew, ximpact[iimpact], yimpact[iimpact]);
                    Segment_flush(&seg_impprobr);
                }
            }
        }

        for ( model = 0; model < PAR_NMODELS; model++ ) {
            if ( omega_stop[model] == CONST_UDP ) omega_stop[model] = CONST_UDM;
                // if necessary, setting angle of reach to no data
        }

        if ( jimpact > 0 ) { // printing summary to file and display:

            fprintf(f_summary, "%i", casename);
            printf("   %i\tOK   ", casename);

            for ( model = 0; model < PAR_NMODELS; model++ ) {
                fprintf(f_summary, "\t%.0f\t%.2f", s_movmax[model], omega_stop[model]);
                printf("\t%.0f\t%.2f", s_movmax[model], omega_stop[model]);
            }
            fprintf(f_summary, "\t%.1f\t%.0f\n", runupmax, jimpact * pow( csz, 2 ));
            printf("\t%.1f\t%.0f\n", runupmax, jimpact * pow( csz, 2 ));

            if ( PAR_BACK == 1 && s_movmax[0] >= PAR_SMIN ) { // for back-calculation:

                rid += 1; // updating case id for back-calculation output
                fprintf(f_backfile, "%i\t%i\t%.1f\t%.3f\n", rid, casename3, s_movmax[0], omega_stop[0] );
                    // writing detail output to file
            }
        }

    } // stopping loop over all cases


// COMPUTING RISK INDICATOR SCORE -------------------------------------------------------------

    nimpact_all = 0; // initializing total number of affected pixels

    for (x=0; x<m; x++) { // loop over all pixels:
        for (y=0; y<n; y++) {

            Segment_get(&seg_aoi, &ctrl_aoi, x, y); // control for area of interest
            Segment_get(&seg_iis, &iisold, x, y); // impact indicator score

            if ( iisold > 0 ) nimpact_all += 1; // updating total number of affected pixels

            if ( PAR_CTRLIMPOBJ == 1 && ctrl_aoi == 1 ) {

                Segment_get(&seg_his, &his_test, x, y); // impact hazard indicator score
                Segment_get(&seg_impobj, &ctrl_impobj, x, y); // impact objects

                if ( ctrl_impobj > 0 && his_test > 0 ) {

                    if ( PAR_CTRLOBJSCORES == 1 ) { // if object scores are available:

                        for ( nimpobj = 0; nimpobj < PAR_NIMPOBJ; nimpobj++ ) // searching for correct impact object id
                            if ( IMP_NAME[nimpobj] == ctrl_impobj ) break; // leaving loop when id has been found

                        ris = (int)( pow( pow( (double)his_test, 1.5 ) * pow ( (double)IMP_S[nimpobj], 1.5 ), (double)1 / (double)3 ));
                            // risk indicator score
                    }
                    else ris = his_test;
                    if ( ris > ris_obj[nimpobj] ) ris_obj[nimpobj] = ris; // updating maximum risk indicator score of object

                    Segment_put(&seg_ris_impobj, &ris, x, y);
                }
            }
        }
    }
    if ( PAR_CTRLIMPOBJ == 1 ) Segment_flush(&seg_ris_impobj);

    if ( PAR_CTRLIMPOBJ == 1 && PAR_CTRLOBJSCORES == 1 ) {
        for ( nimpobj = 0; nimpobj < PAR_NIMPOBJ; nimpobj++ ) // updating risk indicator score file:
            fprintf(f_impobj, "%i %i %i %i", IMP_NAME[nimpobj], IMP_TYPE[nimpobj], IMP_S[nimpobj], ris_obj[nimpobj]);
        fclose(f_impobj);
    }

    if ( PAR_MULT == 1 ) fprintf(f_nout, "%i\n%.1f\n%.3f\n1\n", nimpact_all, s_movmax[0], omega_stop[0] );
        // updating file with information on model run


// FINALIZING PROBABILITIES -------------------------------------------------------------------


    if ( PAR_PROB == 1 ) {

        for (x=0; x<m; x++) { // loop over all pixels:
            for (y=0; y<n; y++) {

                Segment_get(&seg_aoi, &ctrl_aoi, x, y); // area of interest

                if ( ctrl_aoi == 1 && PAR_AFLAG == 1 ) { // average of impact probability:

                    Segment_get(&seg_impquant, &impquantold, x, y); // number of cases possibly affecting the pixel
                    if ( PAR_MULT == 0 ) Segment_get(&seg_relprobr, &relprobrold, x, y); // sum of release probability
                    Segment_get(&seg_impprobr, &impprobrold, x, y); // sum of impact probability
                    if ( impquantold > 0 ) {
                        if ( PAR_MULT == 0 ) relprobrnew = relprobrold / impquantold; // average zonal release probability
                        impprobrnew = impprobrold / impquantold; // average impact probability
                    }
                    else {
                        relprobrnew = 0; // average zonal release probability
                        impprobrnew = 0; // average impact probability
                    }
                    if ( PAR_MULT == 0 ) Segment_put(&seg_relprobr, &relprobrnew, x, y);
                    Segment_put(&seg_impprobr, &impprobrnew, x, y);
                }
                else if ( ctrl_aoi < 1 ) {

                    if ( PAR_MULT == 0 ) relprobrnew = CONST_UDM;
                    impprobrnew = CONST_UDM;
                    if ( PAR_MULT == 0 ) Segment_put(&seg_relprobr, &relprobrnew, x, y);
                    Segment_put(&seg_impprobr, &impprobrnew, x, y);
                }
            }
        }
        if ( PAR_MULT == 0 ) Segment_flush(&seg_relprobr);

        Segment_flush(&seg_impprobr);
    }


// WRITING OUTPUT RASTER MAPS -----------------------------------------------------------------


    if ( PAR_MULT == 0 ) {

        sprintf( miis, "%s_iis", PAR_PREFIX );

        for ( model = 0; model < PAR_NMODELS; model++ ) {
            sprintf( mout, "%s_iM%d", PAR_PREFIX, PAR_MID[model] );
            foutrasti( mout, seg_iM[model] );
        }
        sprintf( msmov, "%s_smov", PAR_PREFIX );
        sprintf( mzmov, "%s_zmov", PAR_PREFIX );
        sprintf( momega, "%s_omega", PAR_PREFIX );
        sprintf( mimpquant, "%s_arz", PAR_PREFIX );

        if ( PAR_CTRLIMPOBJ == 1 ) sprintf( mris_impobj, "%s_iris", PAR_PREFIX );
        if ( PAR_PROB == 1 ) {
            sprintf( mrelprobr, "%s_prz", PAR_PREFIX );
            sprintf( mimpprobr, "%s_pi", PAR_PREFIX );
        }
        if ( PAR_SCORE == 1 ) sprintf( mhis, "%s_ihis", PAR_PREFIX );
    }
    else {
        sprintf( miis, "%s_iis%d", PAR_PREFIX, xint );
        if ( PAR_PROB == 1 ) sprintf( mimpprobr, "%s_pi%d", PAR_PREFIX, xint );
    }

    foutrasti( miis, seg_iis );

    if ( PAR_MULT == 0 ) {

        foutrastd( msmov, seg_smov );
        foutrastd( mzmov, seg_zmov );
        foutrastd( momega, seg_omega );
        foutrasti( mimpquant, seg_impquant );

        if ( PAR_CTRLIMPOBJ == 1 ) foutrasti( mris_impobj, seg_ris_impobj );
        if ( PAR_PROB == 1 ) foutrastd( mrelprobr, seg_relprobr );
        if ( PAR_SCORE == 1 ) foutrasti( mhis, seg_his );
    }
    if ( PAR_PROB == 1 ) foutrastd( mimpprobr, seg_impprobr );


// CLOSING FILES AND RELEASING SEGMENT DATA ---------------------------------------------------


    fclose(f_summary);
    if ( PAR_MULT == 1 ) fclose(f_nout);
    if ( PAR_BACK == 1 ) fclose(f_backfile);

    Segment_release(&seg_aoi);
    Segment_release(&seg_elev);
    if ( PAR_CTRLPROBMAP == 1 ) Segment_release(&seg_probmap);
    if ( PAR_CTRLRELEASEMAP == 1 ) Segment_release(&seg_releasemap);
    if ( PAR_X == 1 ) Segment_release(&seg_releasemap2);
    if ( PAR_CTRLSCOREMAP == 1 ) Segment_release(&seg_scoremap);
    if ( PAR_CTRLIMPOBJ == 1 ) Segment_release(&seg_impobj);
    if ( PAR_BACK == 1 ) Segment_release(&seg_imparea);

    Segment_release(&seg_iis);
    Segment_release(&seg_impquant);

    if ( PAR_MULT == 0 ) {

        for ( model = 0; model < PAR_NMODELS; model++ ) Segment_release(&seg_iM[model]);

        Segment_release(&seg_smov);
        Segment_release(&seg_zmov);
        Segment_release(&seg_omega);

        if ( PAR_CTRLIMPOBJ == 1 ) Segment_release(&seg_ris_impobj);
        if ( PAR_PROB == 1 ) {
            Segment_release(&seg_relprobr);
            Segment_release(&seg_impprob);
        }
        if ( PAR_SCORE == 1 ) Segment_release(&seg_his);
    }
    if ( PAR_PROB == 1 ) Segment_release(&seg_impprobr);


// EXIT ---------------------------------------------------------------------------------------


    printf("\n");
    exit(EXIT_SUCCESS);
}
